import re
from collections import defaultdict, namedtuple
from pathlib import Path

import requests
import yaml

from openpecha.formatters.layers import AnnType, SubText

INFO = "[INFO] {}"


class Serialize(object):
    """
    This class is used when serializing the .opf into anything else (Markdown, TEI, etc.).

    It is relatively abstract and needs to be inherited by a class doing an actual serialization.

    Note that currently we suppose that we're only adding characters, never removing any. This can
    change in the future but let's start simple.

    To use it, instantiate a concrete class with the path of the opf file, and call apply_layers() then get_result()
    """

    def __init__(
        self, opf_path, text_id=None, vol_ids=None, layers=None, index_layer=None
    ):
        self.opf_path = Path(opf_path)
        self.meta = self.get_meta_data()
        self.text_id = text_id
        self.index_layer = index_layer
        self.layers = layers
        self.n_char_shifted = []
        self.text_spans = {}
        self.base_layers = {}
        if self.text_id:
            self.text_spans = self.get_text_spans(text_id)
            self.index_layer = self.get_index_layer(text_id)
            if self.text_spans:
                self.base_layers = self.get_text_base_layer()
        else:
            if not vol_ids:
                vol_ids = [vol.stem for vol in (self.opf_path / "base").iterdir()]
            for vol_id in vol_ids:
                text_spans = {vol_id: {"start": 0, "end": float("inf")}}
                base_layers = {vol_id: self.get_base_layer(vol_id=vol_id)}
                self.text_spans.update(text_spans)
                self.base_layers.update(base_layers)
        """
        The chars_toapply is an important piece of the puzzle here. Basically applying the changes to the string directly is a
        bad idea for several reasons:
          - changing a big string with each annotation is costly
          - the application can be complex as character coordinate keep changing all the time

        So instead of just changing the string, we fill an object with all the characters we're going to add, then
        apply all the changes at once. This simplifies the code, the logic and is more efficient.

        The object has the following structure:

        {
            charcoord: (["string to apply before"],["string to apply after"])
        }

        So for example:
          Annotated text = XXXXXXXXX<title>TTTTTT</title>XXXXXXX
          - there is an annotation that goes from character 10 to character 15
          - the serialization you want to make is to add "<title>" before and "</title>" after the title
          - the chars_toapply object will be:

        {
            10: ([], ["<title>"]),
            15: (["</title>"], [])
        }
        """
        self.chars_toapply = defaultdict(dict)
        # layer lists the layers to be applied, in the order of which they should be applied
        # by convention, when it is None, all layers are applied in alphabetical order (?)

    def get_n_char_shitted(self, end):
        n_shifted = 0
        for pos, n_chars in self.n_char_shifted:
            if end >= pos:
                n_shifted += n_chars
        return n_shifted

    def _get_adapted_span(self, span, vol_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            vol_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = max(0, span["start"] - self.text_spans[vol_id]["start"])
        adapted_end = span["end"] - self.text_spans[vol_id]["start"]
        n_char_shifted = self.get_n_char_shitted(span["start"])
        adapted_start += n_char_shifted
        adapted_end += n_char_shifted
        return adapted_start, adapted_end

    def get_meta_data(self):
        opf_path = self.opf_path
        try:
            meta = yaml.safe_load((opf_path / "meta.yml").open())
        except Exception:
            print("Meta data not Found!!!")
            meta = {}
        return meta

    def load_layer(self, fn):
        return yaml.safe_load(fn.open())

    def get_text_spans(self, text_id):
        """
        get spans of text
        """
        text_span = {}
        index_layer = self.load_layer(self.opf_path / "index.yml")
        for id, anno in index_layer["annotations"].items():
            if anno["parts"]:
                for sub_topic in anno["parts"]:
                    if sub_topic["work_id"] == text_id:
                        text_span[f'v{sub_topic["span"]["vol"]:03}'] = sub_topic["span"]
            if anno["work_id"] == text_id:
                for span in anno["span"]:
                    text_span[f'v{span["vol"]:03}'] = span
        return text_span

    def get_index_layer(self, text_id):
        index_layer = self.load_layer(self.opf_path / "index.yml")
        text_index_layer = defaultdict(str)
        text_index_layer["id"] = index_layer["id"]
        text_index_layer["annotation_type"] = index_layer["annotation_type"]
        text_index_layer["revision"] = index_layer["revision"]
        annotations = defaultdict(str)
        for id, anno in index_layer["annotations"].items():
            if anno["work_id"] == text_id:
                annotations[id] = anno
            elif anno["parts"]:
                annotation = {}
                annotation_span_list = []
                for sub_topic in anno["parts"]:
                    if sub_topic["work_id"] == text_id:
                        annotation["work_id"] = sub_topic["work_id"]
                        annotation_span_list.append(sub_topic["span"])
                        annotation["parts"] = []
                if annotation_span_list:
                    annotation["span"] = annotation_span_list
                    annotations[id] = annotation
        text_index_layer["annotations"] = annotations
        return text_index_layer

    def get_base_layer(self, vol_id=None):
        """
        return text for given span
        """
        if self.text_id:
            vol_base = (self.opf_path / f"base/{vol_id}.txt").read_text()
            start = self.text_spans[vol_id]["start"]
            end = self.text_spans[vol_id]["end"]
            return vol_base[start:end]
        else:
            vol_base = (self.opf_path / f"base/{vol_id}.txt").read_text()
            return vol_base

    def get_text_base_layer(self):
        """
        returns base text of text's volumes: dict

        for example:
        {
            'base/v005': text of given span of v001,
            ....
        }
        """
        base_layers = {}
        for vol_id in self.text_spans:
            base_layers[vol_id] = self.get_base_layer(vol_id)
        return base_layers

    def apply_layer(self, vol_id, layer_id):
        """
        This reads the file opfpath/layers/layer_id.yml and applies all the annotations it contains, in the order in which they appear.
        I think it can be implemented in this class by just calling self.apply_annotation on each annotation of the file.
        """
        layer_fn = self.opf_path / "layers" / vol_id / f"{layer_id}.yml"
        if not layer_fn.is_file():
            return
        layer = yaml.safe_load(layer_fn.open())
        for ann_id, ann in layer["annotations"].items():
            # text begins in middle of the page
            if (
                ann["span"]["end"] >= self.text_spans[vol_id]["start"]
                and ann["span"]["start"] <= self.text_spans[vol_id]["end"]
            ):
                ann["type"] = layer["annotation_type"]
                ann["id"] = ann_id
                try:
                    uuid2localid = layer["local_ids"]
                except Exception:
                    uuid2localid = ""
                self.apply_annotation(vol_id, ann, uuid2localid)

    def apply_index(self):
        for ann_id, topic in self.index_layer["annotations"].items():
            topic_ann = defaultdict(str)
            sub_topics = topic["parts"]
            for i, sub_topic in enumerate(sub_topics):
                if sub_topic:
                    vol_id = f"v{sub_topic['span']['vol']:03}"
                    sub_topic["type"] = AnnType.sub_topic
                    if sub_topic["work_id"] != sub_topics[i - 1]["work_id"]:
                        self.apply_annotation(vol_id, sub_topic)
            if topic["span"]:
                vol_id = f"v{topic['span'][0]['vol']:03}"
                topic_ann["type"] = AnnType.topic
                topic_ann["span"] = topic["span"][0]
                topic_ann["work_id"] = topic["work_id"]
                self.apply_annotation(vol_id, topic_ann)

    def get_all_layer(self, vol_id):
        """
        Returns all the layerid of layer from the layer directory
        """
        return [
            layer.stem
            for layer in (self.opf_path / "layers" / vol_id).iterdir()
            if layer.suffix == ".yml"
        ]

    def apply_layers(self):
        """
        This applies all the layers recorded in self.layers. If self.layers is none, it reads all the layers from the layer directory.
        """
        if not self.index_layer:
            index_path = self.opf_path / "index.yml"
            if index_path.is_file():
                self.index_layer = self.load_layer(index_path)
                self.apply_index()
        else:
            self.apply_index()
        for vol_id in self.base_layers:
            if not self.layers:
                self.layers = self.get_all_layer(vol_id)
            for layer_id in self.layers:
                self.apply_layer(vol_id, layer_id)

    def add_chars(self, vol_id, cc, frombefore, charstoadd):
        """
        This records some characters to add at a character coordinate (cc), either frombefore (from the left) or after. before is a boolean.
        """
        if cc not in self.chars_toapply[vol_id]:
            self.chars_toapply[vol_id][cc] = ([], [])
        if frombefore:  # if from the left, layers should be applied in reverse order
            self.chars_toapply[vol_id][cc][0].insert(0, charstoadd)
        else:
            self.chars_toapply[vol_id][cc][1].append(charstoadd)

    def apply_annotation(self, vol_id, annotation):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            vol_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        raise NotImplementedError(
            "The Serialize class doesn't provide any serialization, please use a subclass such ass SerializeMd"
        )

    def _assign_line_layer(self, result, vol_id):
        def _get_page_index(line):
            page_index = ""
            if re.search("[𰵀-󴉱]", line):
                i = 2
            else:
                i = 1
            while line[i] != "]":
                page_index += line[i]
                i += 1
            return page_index

        result_with_line = ""
        page_index = ""
        n_line = 1
        for line in result.split("\n"):
            if not line:
                continue
            if line[0] == "[" and line[1] != vol_id[0]:
                page_index = _get_page_index(line)
                n_line = 1
            elif line[0] != "[":
                line = f"[{page_index}.{n_line}]" + line
                n_line += 1
            result_with_line += line + "\n"
        return result_with_line

    def get_result(self, line_num=True):
        """
        returns a string which is the base layer where the changes recorded in self.chars_toapply have been applied.

        The algorithm should be something like:
        """
        result = {}
        # don't actually do naive string concatenations
        # see https://waymoot.org/home/python_string/ where method 5 is good
        for vol_id, base_layer in self.base_layers.items():
            cur_vol_result = ""
            # if self.text_id:
            #     cur_vol_result += f"\n[{vol_id}]\n"
            i = 0
            for c in base_layer:
                # UTF bom \ufeff takes the 0th index
                if c == "\ufeff":
                    continue
                if i in self.chars_toapply[vol_id]:
                    apply = self.chars_toapply[vol_id][i]
                    for s in apply[0]:
                        cur_vol_result += s
                    cur_vol_result += c
                    for s in apply[1]:
                        cur_vol_result += s
                else:
                    cur_vol_result += c
                i += 1

            if "Pagination" in self.layers:
                cur_vol_result = self._assign_line_layer(cur_vol_result, vol_id)
            result.update({vol_id: cur_vol_result})
        return result



import re


class StringChunker:
    def __init__(self, full_string, pattern, n):
        self.pattern = pattern
        self.string = full_string
        self.n = n

    @staticmethod
    def get_n_char(string, n):
        """
        Returns n unicode char
        """
        return string[:n]

    def find_last_match_pos(self, string):
        """
        Finds and return the last possible match of the pattern
        """
        match = None
        for match in re.finditer(self.pattern, string):
            pass
        return match.end() if match else -1

    def find_first_match_pos(self, string):
        """
        Finds and return the first possible match of the pattern
        """
        match = None
        for match in re.finditer(self.pattern, string):
            return match.end() if match else -1

    def get_chunk(self, string, n):
        """
        Returns an index of where the string has the last char, if the char is not found in n char,
        the function doubles n and looks for the first char in that double chunk,
        if the char is still not found it looks for the last space in the original chunk,
        if still no space was found in the chunk than just returns n to cut up the string arbitrarily
        """
        chunk = self.get_n_char(string, n)
        index = self.find_last_match_pos(chunk)
        if index > -1:
            return index
        else:
            double_chunk = self.get_n_char(string, n * 2)
            index = self.find_last_match_pos(double_chunk)
            if index > -1:
                return index
            else:
                index = chunk.rfind(" ")
                if index > -1:
                    return index + 1
                else:
                    return n

    def get_chunks(self):
        """
        Returns a list of indexes of the string cut up in chunks of a maximum of n length and cut up at the last
        instance of regex
        """
        list_of_index = [0]
        string = self.string
        while len(string[list_of_index[-1] :]) > self.n:
            index = self.get_chunk(string[list_of_index[-1] :], self.n)
            list_of_index.append(list_of_index[-1] + index)

        list_of_index.append(list_of_index[-1] + len(string[list_of_index[-1] :]))
        return list_of_index



from pathlib import Path

import yaml
from git import Repo

from openpecha.buda.op import Openpecha


class OpenpechaBare(Openpecha):
    """
    A class inheriting from the Openpecha class that defines the path to the local bare git repo.

    TODO: conceptually there's really a lot in common with op_fs.py as it's just reading an OPF folder
          in a git repo, but the implementation is completely different because the way to access files
          in a bare git repo is very specific. This could also be considered a serializer, like op_fs.py.
    """

    def __init__(self, lname, path=None, repo=None, rev="HEAD"):
        self.rev = rev
        if repo is not None:
            self.repo = repo
        else:
            self.repo = Repo(path)
        Openpecha.__init__(self, lname)

    def get_rev(self):
        return self.rev

    def read_file_content(self, oppath):
        return self.repo.git.show(f"{self.rev}:{self.lname}.opf/" + oppath)

    def read_file_content_yml(self, oppath):
        ymlstr = self.repo.git.show(f"{self.rev}:{self.lname}.opf/" + oppath)
        return yaml.safe_load(ymlstr)

    def list_paths(self):
        """
        Getting all the files in the bare repo
        """
        files = self.repo.git.ls_tree(r=self.rev).split("\n")
        # removing the xxx.opf at the beginning
        files = [file.split("\t")[-1] for file in files]
        files = [
            file[len(self.lname) + 5 :]
            for file in files
            if file.startswith(self.lname + ".")
            and (file.endswith(".txt") or file.endswith(".yml"))
        ]
        return files



from pathlib import Path


def _mkdir(path):
    if path.is_dir():
        return path
    path.mkdir(exist_ok=True, parents=True)
    return path


# Path
BASE_PATH = _mkdir(Path.home() / ".openpecha")
OPF_OUTPUT_PATH = _mkdir(BASE_PATH / "opfs")

# Pecha
PECHA_PREFIX = "P"
N_SIG = 6  # no. of significant digits for pecha id



import pathlib
from os import walk

import yaml

from openpecha.buda.op import Openpecha


class OpenpechaFS(Openpecha):
    """
    A class inheriting from the Openpecha class that defines the path to the local .opf folder

    TODO: perhaps this should be renamed OPF and be moved into serializers/ as it's really a reader/serializer for an .opf folder

    TODO: we could also implement write functions. Basically there would be:
       - readFromPath(path): would more or less replace the current __init__
       - writeToPath(path): would write the content to a path (removing all files that used to be there)
    """

    def __init__(self, lname, path_to_opf):
        Openpecha.__init__(self, lname)
        self.path = path_to_opf

    def get_rev(self):
        return None

    def list_paths(self):
        """
        Getting all the files in the directory

        TODO: this is probably wrong, there should be a unit test for it
        """
        files = []
        for (dirpath, dirnames, filenames) in walk(self.path):
            for file in filenames:
                files.append(f"{dirpath}/{file}".replace(self.path, ""))

        return files

    def read_file_content(self, oppath):
        with open(f"{self.path}/" + oppath) as f:
            return f.read()

    def read_file_content_yml(self, oppath):
        with open(f"{self.path}/" + oppath) as file:
            return yaml.safe_load(file)



from openpecha.buda.chunker import StringChunker


class TibetanEasyChunker(StringChunker):
    def __init__(self, full_string, n):
        StringChunker.__init__(
            self,
            full_string,
            "([སའངགདནབམརལཏ]\u0f7c།[^ཀ-ཬ]*|།།[^ཀ-ཬ༠-༩]*།[^ཀ-ཬ༠-༩]*)",
            n,
        )



import os
import shutil
import time
from pathlib import Path
from uuid import uuid4

from git import Repo
from github import Github

org = None


def _get_openpecha_org(org_name, token):
    """OpenPecha github org singleton."""
    global org
    if org is None:
        g = Github(token)
        org = g.get_organization(org_name)
    return org


def get_github_repo(repo_name, org_name, token):
    org = _get_openpecha_org(org_name, token)
    repo = org.get_repo(repo_name)
    return repo


def create_github_repo(path, org_name, token):
    org = _get_openpecha_org(org_name, token)
    repo = org.create_repo(path.name)
    time.sleep(2)
    return repo._html_url.value


def commit(repo, message, not_includes, branch="master"):
    has_changed = False

    # add untrack fns
    for fn in repo.untracked_files:

        ignored = False
        for not_include_fn in not_includes:
            if not_include_fn in fn:
                ignored = True

        if ignored:
            continue

        if fn:
            repo.git.add(fn)
        if has_changed is False:
            has_changed = True

    # add modified fns
    if repo.is_dirty() is True:
        for fn in repo.git.diff(None, name_only=True).split("\n"):
            if fn:
                repo.git.add(fn)
            if has_changed is False:
                has_changed = True

    if has_changed is True:
        if not message:
            message = "Initial commit"
        repo.git.commit("-m", message)
        repo.git.push("origin", branch)


def create_local_repo(path, remote_url, org, token):
    if (path / ".git").is_dir():
        return Repo(path)
    else:
        repo = Repo.init(path)
        old_url = remote_url.split("//")
        auth_remote_url = f"{old_url[0]}//{org}:{token}@{old_url[1]}"
        repo.create_remote("origin", auth_remote_url)

        repo.config_writer().set_value("user", "name", "10zinten").release()
        repo.config_writer().set_value("user", "email", "ten13zin@gmail.com").release()
        return repo


def create_orphan_branch(repo, branch_name):
    repo.git.checkout("master")
    repo.git.checkout("--orphan", branch_name)

    # move base-text root level
    repo_path = Path(repo.working_dir)
    pecha_opf_path = repo_path / f"{repo_path.name}.opf"
    for vol_fn in (pecha_opf_path / "base").iterdir():
        shutil.move(str(vol_fn), str(repo_path))

    repo.git.rm("-rf", str(pecha_opf_path.name))
    repo.git.rm("-f", "README.md")


def github_publish(
    path,
    message=None,
    not_includes=None,
    layers=[],
    org="OpenPecha",
    token=os.environ.get("GITHUB_TOKEN"),
):
    path = Path(path)
    remote_repo_url = create_github_repo(path, org, token)
    local_repo = create_local_repo(path, remote_repo_url, org, token)
    commit(local_repo, message, not_includes)

    for layer in layers:
        create_orphan_branch(local_repo, layer)
        commit(local_repo, message, not_includes, branch=layer)


def create_file(
    repo_name,
    path,
    content,
    msg,
    update=False,
    org="OpenPecha",
    token=os.environ.get("GITHUB_TOKEN"),
):
    repo = get_github_repo(repo_name, org, token)
    if update:
        old_content = repo.get_contents(path)
        repo.update_file(old_content.path, msg, content, old_content.sha)
    else:
        repo.create_file(path, msg, content)


def get_bumped_tag(repo):
    try:
        latest_release_tag = repo.get_latest_release().tag_name
    except Exception:
        return "v0.1"

    tag_number = float(latest_release_tag[1:])
    bump_tag_number = round(tag_number + 0.1, 1)
    return f"v{bump_tag_number}"


def _add_tag_in_filename(path, tag_name):
    path = Path(path)
    old_name = path.stem
    old_extension = path.suffix
    directory = path.parent
    new_name = f"{old_name}-{tag_name}{old_extension}"
    target_path = directory / new_name
    path.rename(target_path)
    return target_path


def upload_assets(release, tag_name=None, asset_paths=[]):
    if not tag_name:
        tag_name = release.tag_name
    download_url = ""
    for asset_path in asset_paths:
        # asset_path = _add_tag_in_filename(asset_path, tag_name)
        asset = release.upload_asset(str(asset_path))
        download_url = asset.browser_download_url
        print(f"[INFO] Uploaded asset {asset_path}")
    return download_url


def create_release(
    repo_name,
    prerelease=False,
    asset_paths=[],
    org="OpenPecha",
    token=os.environ.get("GITHUB_TOKEN"),
):
    repo = get_github_repo(repo_name, org, token)
    if prerelease:
        bumped_tag = uuid4().hex
        message = "Pre-release for download"
    else:
        bumped_tag = get_bumped_tag(repo)
        message = "Official Release"
    new_release = repo.create_git_release(
        bumped_tag, bumped_tag, message, prerelease=prerelease
    )
    print(f"[INFO] Created release {bumped_tag} for {repo_name}")
    asset_download_url = upload_assets(
        new_release, tag_name=bumped_tag, asset_paths=asset_paths
    )
    return asset_download_url


def add_assets_to_latest_release(repo_name, asset_paths=[]):
    repo = get_github_repo(repo_name)
    lastest_release = repo.get_latest_release()
    upload_assets(lastest_release, asset_paths=asset_paths)


def create_readme(metadata, path):
    result = ""
    # add title
    result += f"## Title\n\t- {metadata.get('title')}\n\n"

    # add author
    result += f"## Author\n\t- {metadata.get('author')}\n\n"

    readme_fn = path / "README.md"
    readme_fn.write_text(result)


def delete_repo(repo_name):
    org = _get_openpecha_org()
    repo = org.get_repo(repo_name)
    repo.delete()


if __name__ == "__main__":
    asset_download_url = create_release(
        "P000780", prerelease=True, asset_paths=Path("assets").iterdir()
    )
    print(asset_download_url)
    create_release("P000780", asset_paths=Path("assets").iterdir())



import base64
import logging
import os
from abc import ABC, abstractclassmethod

import requests
from github import Github

from .config import GITHUB_BUCKET_CONFIG
from .utils import create_pecha_id


class Bucket(ABC):
    """A class representing a Bucket on Cloud Storage."""

    def __init__(self, name, config):
        self.name = name
        self.config = config
        super().__init__()

    @abstractclassmethod
    def _get_client(self, name, token):
        pass

    @abstractclassmethod
    def get_pecha_base(self, id):
        pass

    @abstractclassmethod
    def get_all_pechas_base(self):
        pass


class GithubBucket(Bucket):
    def __init__(self, name, config=GITHUB_BUCKET_CONFIG):
        self.client = self._get_client(name, config["token"])
        super().__init__(name, config)

    def _get_client(self, name, token):
        """Return github org object"""
        g = Github(token)
        org = g.get_organization(name)
        return org

    @staticmethod
    def get_blob_content(blob_url):
        """
        Get the blob content from the url provided in parameter
        """
        base64_content = requests.get(blob_url).json()["content"]
        return base64_content

    @staticmethod
    def decode_bas64_blob(base64_blob):
        """
         Decode the base64 encoded blob into UTF-8
        """
        decoded_list_data = base64.b64decode(base64_blob).decode("utf-8")
        return decoded_list_data

    def _get_content(self, gh_file_obj):
        blob_content = self.get_blob_content(gh_file_obj.git_url)
        decoded_blob = self.decode_bas64_blob(blob_content)
        return decoded_blob

    def get_pecha_base(self, id):
        try:
            repo = self.client.get_repo(id)
        except Exception:
            return [], ""
        for vol_base_file_obj in repo.get_contents(f"{id}.opf/base"):
            vol_base_content = self._get_content(vol_base_file_obj)
            logging.info(f"Downloaded {vol_base_file_obj.name}")
            yield vol_base_content, vol_base_file_obj.name

    def get_all_pechas_base(self):
        for pecha_num in range(
            self.config["catalog"]["start_id"], self.config["catalog"]["end_id"] + 1
        ):
            pecha_id = create_pecha_id(pecha_num)
            logging.info(f"Downloading {pecha_id}")
            yield pecha_id, self.get_pecha_base(pecha_id)



import os
import re
from os import stat
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from openpecha.formatters.layers import AnnType
from openpecha.formatters.tsadra import TsadraTemplate

from .serialize import Serialize


class Tsadra_template:
    """Variables are important components of tsadra epub template."""

    # SP = Start_Payload
    span_EP = "</span>"
    para_EP = "</p>"
    ft = '<span class="front-title">'
    cover_page_book_title_SP = '<span class="credits-page_front-title">'
    book_title_SP = '<span class="tibetan-book-title">'
    book_number_SP = f'<p class = "credits-page_front-page---book-number">{ft}'
    author_SP = '<p class="credits-page_front-page---text-author"><span class="front-page---text-titles">'
    chapter_SP = '<span class="tibetan-chapter">'
    tsawa_SP = '<span class="tibetan-root-text">'
    tsawa_verse_SP = '<span class="tibetan-root-text_tibetan-root-text-middle-lines tibetan-root-text">'
    quatation__verse_SP = (
        '<span class="tibetan-citations-in-verse_tibetan-citations-middle-lines">'
    )
    quatation__SP = '<span class="tibetan-external-citations">'
    sabche_SP = '<span class="tibetan-sabche1">'
    yigchung_SP = '<span class="tibetan-commentary-small">'


class EpubSerializer(Serialize):
    """Epub serializer class for OpenPecha."""

    def __get_adapted_span(self, span, vol_id):
        """Adapts the annotation span to base-text of the text

        Adapts the annotation span, which is based on volume base-text
        to text base-text.

        Args:
            span (dict): span of a annotation, eg: {start:, end:}
            vol_id (str): id of vol, where part of the text exists.

        Returns:
            adapted_start (int): adapted start based on text base-text
            adapted_end (int): adapted end based on text base-text

        """
        adapted_start = span["start"] - self.text_spans[vol_id]["start"]
        adapted_end = span["end"] - self.text_spans[vol_id]["start"]
        return adapted_start, adapted_end

    def apply_annotation(self, vol_id, ann, uuid2localid):
        """Applies annotation to specific volume base-text, where part of the text exists.

        Args:
            vol_id (str): id of vol, where part of the text exists.
            ann (dict): annotation of any type.

        Returns:
            None

        """
        only_start_ann = False
        start_payload = "("
        end_payload = ")"
        # if ann["type"] == AnnType.pagination:
        #     start_payload = f'[{ann["page_index"]}] {ann["page_info"]}\n'
        #     only_start_ann = True
        if ann["type"] == AnnType.correction:
            start_payload = "("
            end_payload = f',{ann["correction"]})'
        elif ann["type"] == AnnType.peydurma:
            start_payload = "#"
            only_start_ann = True
        elif ann["type"] == AnnType.error_candidate:
            start_payload = "["
            end_payload = "]"
        elif ann["type"] == AnnType.book_title:
            try:
                if ann["iscover"]:
                    start_payload = Tsadra_template.cover_page_book_title_SP
                else:
                    start_payload = Tsadra_template.book_title_SP
            except Exception:
                start_payload = Tsadra_template.book_title_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.book_number:
            start_payload = Tsadra_template.book_number_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann["type"] == AnnType.author:
            start_payload = Tsadra_template.author_SP
            end_payload = f"{Tsadra_template.span_EP}{Tsadra_template.para_EP}"
        elif ann["type"] == AnnType.chapter:
            start_payload = Tsadra_template.chapter_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.tsawa:
            try:
                if ann["isverse"]:
                    start_payload = Tsadra_template.tsawa_verse_SP
                else:
                    start_payload = Tsadra_template.tsawa_SP
            except Exception:
                start_payload = Tsadra_template.tsawa_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.citation:
            try:
                if ann["isverse"]:
                    start_payload = Tsadra_template.quatation__verse_SP
                else:
                    start_payload = Tsadra_template.quatation__SP
            except Exception:
                start_payload = Tsadra_template.quatation__SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.sabche:
            start_payload = Tsadra_template.sabche_SP
            end_payload = Tsadra_template.span_EP
        elif ann["type"] == AnnType.yigchung:
            start_payload = Tsadra_template.yigchung_SP
            end_payload = Tsadra_template.span_EP

        start_cc, end_cc = self.__get_adapted_span(ann["span"], vol_id)
        self.add_chars(vol_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(vol_id, end_cc, False, end_payload)

    def get_syls(self, vol_text):
        result = []
        cur_syl = ""
        syls = re.split("(་|།)", vol_text)
        for i, syl in enumerate(syls):
            if i % 2 == 0:
                cur_syl += syl
            else:
                cur_syl += syl
                result.append(cur_syl)
                cur_syl = ""
        return result

    def update_line_break(self, result_text):
        result_text = result_text.replace("།\n", "། ")
        result_text = result_text.replace("\n", "")
        syls = self.get_syls(result_text)
        walker = 0
        result = []
        br_flag = False
        for syl in syls:
            if br_flag:
                if "།" in syl:
                    result.append(syl)
                    result.append("<br>")
                    br_flag = False
                    walker = 0
                else:
                    result.append(syl)
            else:
                if walker == 499:
                    br_flag = True
                result.append(syl)
            walker += 1
        return "".join(result)

    def serialize(self, output_path="./output/epub_output"):
        """ This module serialize .opf file to other format such as .epub etc. In case of epub,
        we are using calibre ebook-convert command to do the conversion by passing our custom css template
        and embedding our custom font. The converted output will be then saved in current directory
        as {pecha_id}.epub.

        Args:
        pecha_id (string): Pecha id that needs to be exported in other format

        """
        output_path = Path(output_path)
        pecha_id = self.opf_path.name.split(".")[0]
        out_html_fn = f"{pecha_id}.html"
        try:
            pecha_title = self.meta["ebook_metadata"]["title"]
        except KeyError:
            pecha_title = ""
        try:
            cover_image = self.meta["ebook_metadata"]["cover"]
        except KeyError:
            cover_image = ""
        self.layers = [layer for layer in self.layers if layer != "Pagination"]
        results = self.get_result()
        for vol_id, result in results.items():
            result = self.update_line_break(result)
            results = (
                f"<html>\n<head>\n\t<title>{pecha_title}</title>\n</head>\n<body>\n"
            )
            results += f"{result}</body>\n</html>"
            Path(out_html_fn).write_text(results)
            # Downloading css template file from ebook template repo and saving it
            template = requests.get(
                "https://raw.githubusercontent.com/OpenPecha/ebook-template/master/tsadra_template.css"
            )
            Path("template.css").write_bytes(template.content)
            # click.echo(template.content, file=open('template.css', 'w'))
            # Running ebook-convert command to convert html file to .epub (From calibre)
            # XPath expression to detect chapter titles.
            chapter_Xpath = "//*[@class='tibetan-chapter']"
            font_family = "Monlam Uni Ouchan2"
            font_size = 15
            chapter_mark = "pagebreak"
            cover_path = self.opf_path / f"asset/image/{cover_image}"
            out_epub_fn = output_path / f"{pecha_id}.epub"
            os.system(
                f'ebook-convert {out_html_fn} {out_epub_fn} --extra-css=./template.css --chapter={chapter_Xpath} --chapter-mark="{chapter_mark}" --base-font-size={font_size} --embed-font-family="{font_family}" --cover={cover_path}'
            )
            # Removing html file and template file
            os.system(f"rm {out_html_fn}")
            os.system("rm template.css")

            return out_epub_fn



import logging
import random

from bonltk.text_quality import non_words_ratio


def _sample_text_chunk(text):
    start_idx = random.randint(0, 100000)
    text_chunk = text[start_idx : start_idx + 1000]
    first_syl_idx = text_chunk.find("་")
    last_syl_idx = text_chunk.rfind("་")
    return text_chunk[first_syl_idx + 1 : last_syl_idx]


def is_text_good_quality(text, strategy=non_words_ratio, threshold=0.2):
    text_chunk = _sample_text_chunk(text)
    ratio = strategy(text_chunk)
    text_quality = 1 - ratio
    logging.info(f"Text quality: {text_quality}")
    if text_quality <= threshold:
        return True
    else:
        return False



import logging
import urllib.request
from pathlib import Path
from urllib.error import HTTPError

import requests
from git import Repo

from openpecha.buda.errors import Error
from openpecha.buda.op_bare import OpenpechaBare

ENV = {"GIT_TERMINAL_PROMPT": "0"}


class OpenpechaGit:
    """
    The OpenpechaGit class is here to help manage and pull all the Openpecha objects and store them on a temporary local
    repository
    It is initiated with
    - The Openpecha lname
    - The local dir (by default stored in ~/.openpecha/data)
    - The remote git of Openpecha
    - If we want the bare repo or a working tree on the local branch (by default the bare repo to save some space)
    """

    def __init__(
        self,
        op_lname,
        cache_dir=str(Path.home() / ".cache" / "openpecha"),
        openpecha_dstgit="https://github.com/OpenPecha",
        bare=True,
    ):
        self.lname = op_lname
        self.cache_dir = cache_dir
        Path(cache_dir).mkdir(parents=True, exist_ok=True)
        self.openpecha_dstgit = f"{openpecha_dstgit}/{self.lname}.git"
        self.bare = bare
        self.repo = None
        self.synced = False
        self.openpecha = None
        self.openpecharev = None

    def get_repo(self, dst_sync=False):
        """
        gets the repo, sync with dst (Github) if parameter is true, otherwise just return the local repo
        """
        if self.repo is not None:
            if dst_sync and not self.synced:
                self.repo.git.fetch()
                self.synced = True
            return self.repo
        if not self.poti_localgit_exists():
            if dst_sync:
                self.clone()
                return self.repo
            return None
        self.repo = Repo(str(Path(self.cache_dir, self.lname)))
        if dst_sync and not self.synced:
            self.repo.git.fetch()
            self.synced = True
        return self.repo

    def clone(self):
        """
        Given a op_lname, clones the repo from openpecha to self.cache_dir
        """
        if self.poti_localgit_exists():
            return
        self.repo = Repo.clone_from(
            self.openpecha_dstgit,
            str(Path(self.cache_dir, self.lname)),
            env=ENV,
            bare=self.bare,
        )
        self.synced = True

    def poti_localgit_exists(self):
        """
        Check if self.lname has already been cloned by looking if there is
        a folder with the self.lname name in the local directory
        """
        path = Path(self.cache_dir, self.lname)
        return path.is_dir()

    def poti_dstgit_exists(self):
        """
        Test if the self.lname has a corresponding git in self.openpecha_dstgit
        """
        try:
            return urllib.request.urlopen(self.openpecha_dstgit).getcode()
        except HTTPError:
            Error(
                HTTPError,
                f"Poti _{self.lname}_ is not present on the openpecha git: {self.openpecha_dstgit}",
            )

    def get_local_latest_commit(self, dst_sync=False, branchname="master"):
        """
        get the commit to sync to BUDA: the latest tag, or the latest commit of the master branch
        """
        repo = self.get_repo(dst_sync)
        tags = sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
        if tags:
            return tags[-1].commit.hexsha
        try:
            return repo.commit(branchname).hexsha
        except Exception:
            logging.error("cannot find branch %s for %s", branchname, self.lname)
            return None

    def get_openpecha(self, rev=None):
        if rev is None:
            rev = self.get_local_latest_commit()
        if rev is None:
            return None
        if self.openpecha is not None and self.openpecharev == rev:
            return self.openpecha
        self.openpecha = OpenpechaBare(self.lname, repo=self.get_repo(), rev=rev)
        self.openpecharev = rev
        return self.openpecha



import re
from pathlib import Path

from openpecha.buda.op_bare import OpenpechaBare
from openpecha.buda.op_fs import OpenpechaFS
from openpecha.buda.tibetan_easy_chunker import TibetanEasyChunker
from openpecha.serializers.rdf_setup import *


class Rdf:
    """
    TODO:
    - rename in BUDARDFSerializer
    - initialize with an OpenPecha instead of a path
    """

    def __init__(self, oplname, openpecha):
        self.oplname = oplname
        self.lname = f"IE0OP{self.oplname}"
        self.graph_r = bdg[self.lname]
        self.lod_ds = rdflib.Dataset()
        self.lod_g = self.lod_ds.graph(self.graph_r)
        self.lod_g.namespace_manager = nsm
        self.openpecha = openpecha

    def add_triple(self, rdf_subject, rdf_predicate, rdf_object):
        self.lod_g.add((rdf_subject, rdf_predicate, rdf_object))

    def get_graph(self):
        self.set_instance()
        return self.lod_g

    """
    Entry point to build the RDF graph
    """

    def set_instance(self):
        self.add_triple(bdr[f"{self.lname}"], rdf.type, bdo["EtextInstance"])
        meta = self.openpecha.get_meta()
        if "source_metadata" in meta:
            sour = meta["source_metadata"]["id"].split(":")
            if sour[0] == "bdr":
                self.add_triple(
                    bdr[self.lname], bdo["instanceReproductionOf"], bdr["M" + sour[-1]]
                )
                self.add_triple(
                    bdr[self.lname], bdo["contentMethod"], bdr['ContentMethod_OCR']
                )
                self.add_triple(
                    bdr["M" + sour[-1]], bdo["instanceHasReproduction"], bdr[self.lname]
                )
                self.add_triple(
                    bdr[self.lname], bdo["instanceReproductionOf"], bdr[sour[-1]]
                )
                self.add_triple(
                    bdr[sour[-1]], bdo["instanceHasReproduction"], bdr[self.lname]
                )
        self.get_base_volumes()
        self.set_adm()

    def get_base_volumes(self):
        for volume_name in self.openpecha.list_base():
            self.set_etext_asset(volume_name)
            self.add_triple(
                bdr[self.lname],
                bdo["instanceHasVolume"],
                bdr[f"VL{self.lname}_{volume_name}"],
            )
            self.set_etext_ref(volume_name)
            self.set_etext(volume_name)

    def set_etext_asset(self, volume_name):
        volume_basename = f"{self.lname}_{volume_name}"
        volume_number = int(re.search(r"\d+", volume_name).group())
        subject = bdr[f"VL{volume_basename}"]

        self.add_triple(subject, rdf.type, bdo["VolumeEtextAsset"])
        self.add_triple(subject, bdo["volumeHasEtext"], bdr[f"ER{volume_basename}"])
        self.add_triple(
            subject, bdo["volumeNumber"], Literal(volume_number, datatype=XSD.integer)
        )
        self.add_triple(subject, bdo["volumeOf"], bdr[f"{self.lname}"])

    def set_etext_ref(self, volume_name):
        volume_basename = f"{self.lname}_{volume_name}"
        subject = bdr[f"ER{volume_basename}"]

        self.add_triple(subject, rdf.type, bdo["EtextRef"])
        self.add_triple(subject, bdo["eTextResource"], bdr[f"UT{volume_basename}"])
        self.add_triple(subject, bdo["seqNum"], Literal(1, datatype=XSD.integer))

    def set_etext(self, volume_name):
        volume_basename = f"{self.lname}_{volume_name}"
        volume_number = int(re.search(r"\d+", volume_name).group())
        subject = bdr[f"UT{volume_basename}"]

        self.add_triple(subject, rdf.type, bdo["Etext"])
        self.add_triple(subject, bdo["eTextInInstance"], bdr[self.lname])
        self.add_triple(
            subject, bdo["eTextIsVolume"], Literal(volume_number, datatype=XSD.integer)
        )
        self.add_triple(
            subject,
            rdfs.seeAlso,
            Literal(
                f"https://github.com/OpenPecha/{self.oplname}/", datatype=XSD.anyURI
            ),
        )
        self.set_etext_pages(volume_name)
        self.set_etext_chunks(volume_name)

    def set_etext_pages(self, volume_name):
        player = self.openpecha.get_layer(volume_name, "pagination")
        if player is None:
            return
        if "pagination" in player:
            player = player["pagination"]
        if "annotations" not in player:
            return
        for annotation in player["annotations"]:
            self.set_etext_page(annotation, volume_name)

    def set_etext_page(self, annotation, volume_name):
        volume_basename = f"{self.lname}_{volume_name}"
        subject = bdr[f'EP{annotation["id"]}']
        sequence = self.get_sequence(annotation["page_index"])
        start = annotation["span"]["start"]
        end = annotation["span"]["end"]
        self.add_triple(subject, rdf.type, bdo["EtextPage"])
        self.add_triple(subject, bdo["seqNum"], Literal(sequence, datatype=XSD.integer))
        self.add_triple(
            subject, bdo["sliceEndChar"], Literal(end, datatype=XSD.integer)
        )
        self.add_triple(
            subject, bdo["sliceStartChar"], Literal(start, datatype=XSD.integer)
        )
        self.add_triple(bdr[f"UT{volume_basename}"], bdo["eTextHasPage"], subject)

    @staticmethod
    def get_sequence(page_index):
        number = int(re.search(r"\d+", page_index).group())
        return number * 2 if page_index[-1] == "b" else (number * 2) - 1

    def set_etext_chunks(self, volume_name):
        volume_string = self.openpecha.get_base(volume_name)
        chunk_indexes = self.get_chunk_index(volume_string)

        for i in range(0, len(chunk_indexes) - 2):
            self.set_etext_chunk(
                i, chunk_indexes[i], chunk_indexes[i + 1], volume_name, volume_string
            )

    def set_etext_chunk(self, i, start_char, end_char, volume_name, volume_string):
        volume_basename = f"{self.lname}_{volume_name}"
        etext = f"UT{volume_basename}"
        subject = bdr[f"UT{volume_basename}_{int(i):05}"]

        self.add_triple(subject, rdf.type, bdo["EtextChunk"])
        self.add_triple(
            subject,
            bdo["chunkContents"],
            Literal(volume_string[start_char:end_char], lang="bo"),
        )
        self.add_triple(
            subject, bdo["sliceEndChar"], Literal(end_char, datatype=XSD.integer)
        )
        self.add_triple(
            subject, bdo["sliceStartChar"], Literal(start_char, datatype=XSD.integer)
        )

        self.add_triple(bdr[etext], bdo["eTextHasChunk"], subject)

    def set_adm(self):
        subject_r = bda[self.lname]
        einst_r = bdr[self.lname]
        graph_r = bdg[self.lname]
        rev = self.openpecha.get_rev()

        self.add_triple(subject_r, rdf.type, adm["AdminData"])
        self.add_triple(subject_r, adm["adminAbout"], einst_r)
        self.add_triple(subject_r, adm["graphId"], graph_r)
        self.add_triple(subject_r, adm["syncAgent"], bdr["SAOPT"])
        self.add_triple(subject_r, adm["metadataLegal"], bda["LD_BDRC_CC0"])
        self.add_triple(subject_r, adm["gitRevision"], Literal(rev))
        self.add_triple(subject_r, adm["status"], bda["StatusReleased"])
        self.add_triple(subject_r, adm["access"], bda["AccessOpen"])

    @staticmethod
    def get_chunk_index(string):
        chunker = TibetanEasyChunker(string, 1500)
        indexes = chunker.get_chunks()

        return indexes

    """
    Getting details of the rdf
    """

    def print_rdf(self):
        print(self.lod_g.serialize(format="ttl").decode("utf-8"))

    def rdf(self):
        return self.lod_g



import pathlib

import yaml


class Openpecha:
    """
    The Openpecha class that correspond to the https://github.com/OpenPecha/openpecha-catalog catalog.

    TODO: for this class and its implementations, function signatures should be changed a bit:
       - get_layer_list(basefname): lists all the layers in the directory corresponding to basename (and cache it)
       - get_layer(basefname, layerfname): get the content of the layer, put it in the object if not presen
       - get_base_list(): get an array of all the layers fnames (and cache it)
       - get_base(basefname): get the string of base layer (and cache it)
    """

    def __init__(self, lname):
        """
        Initializing the openpecha object with it's name from https://github.com/OpenPecha/openpecha-catalog
        Commit is the commit we want, by default the last commit on master
        """
        self.lname = lname
        self.meta = None
        self.bases = {}
        self.layers = {}
        self.components = self.read_components()

    def read_layer(self, basename, layername):
        return self.read_file_content_yml(
            "layers/" + basename + "/" + layername + ".yml"
        )

    def read_base(self, basename):
        return self.read_file_content("base/" + basename + ".txt")

    def read_meta(self):
        """
        Getting the meta.yml
        """
        return self.read_file_content_yml("meta.yml")

    def get_meta(self):
        if self.meta is not None:
            return self.meta
        self.meta = self.read_meta()
        return self.meta

    def is_ocr(self):
        meta = self.get_meta()
        if "source_metadata" in meta:
            sour = meta["source_metadata"]["id"].split(":")
            if sour[0] == "bdr":
                return True
        return False

    def get_base(self, basename):
        if basename in self.bases:
            return self.bases[basename]
        if basename not in self.components["base"]:
            return None
        self.bases[basename] = self.read_base(basename)
        return self.bases[basename]

    def get_layer(self, basename, layername):
        if basename in self.layers and layername in self.layers[basename]:
            return self.layers[basename][layername]
        if basename not in self.components["base"]:
            return None
        if layername not in self.components["layers"][basename]:
            return None
        if basename not in self.layers:
            self.layers[basename] = {}
        self.layers[basename][layername] = self.read_layer(basename, layername)
        return self.layers[basename]

    def has_layer(self, basename, layername):
        return layername in self.components["layers"][basename]

    def list_layers(self, basename):
        if basename not in self.components["layers"]:
            return None
        return self.components["layers"][basename]

    def list_base(self):
        return self.components["base"]

    def read_components(self):
        """
        get all bases and layers
        """
        paths = self.list_paths()
        res = {"base": [], "layers": {}}

        for f in sorted(paths):
            path = f.split("/")
            if len(path) > 1:
                if path[-2] == "base":
                    basename = pathlib.Path(path[-1]).stem
                    res["base"].append(basename)
                else:
                    basename = path[-2]
                    layername = pathlib.Path(path[-1]).stem
                    if basename not in res["layers"]:
                        res["layers"][basename] = []
                    res["layers"][basename].append(layername)
        return res



import json
import sys
from urllib import parse, request


class Error:
    def __init__(self, type, content):
        self.type = type
        self.content = content
        self.print_error()

    def print_error(self):
        print(self.content, file=sys.stderr)



from pathlib import Path

from ..formatters.layers import AnnType
from ..utils import Vol2FnManager
from .serialize import Serialize


class HFMLSerializer(Serialize):
    """
    HFML (Human Friendly Markup Language) serializer class for OpenPecha.
    """

    def get_local_id(self, ann, uuid2localid):
        try:
            return chr(uuid2localid[ann["id"]])
        except Exception:
            return ""

    def apply_annotation(self, vol_id, ann, uuid2localid=None):
        only_start_ann = False
        start_payload = "("
        end_payload = ")"
        side = "ab"
        local_id = self.get_local_id(ann, uuid2localid)
        if ann["type"] == AnnType.pagination:
            if ann["page_index"] == "0b":
                pg_n = ann["reference"][5:-1]
                pg_side = ann["reference"][-1]
                if "-" in pg_n:
                    pg_n = int(pg_n.split("-")[0])
                    pg_side = side[int(pg_side)]
                    start_payload = f"[{local_id}{pg_n}{pg_side}]"
                else:
                    pg_n = int(pg_n)
                    if pg_side.isdigit():
                        pg_n = str(pg_n) + pg_side
                        pg_side = ""
                    start_payload = f"[{local_id}{pg_n}{pg_side}]"
            else:
                start_payload = f'[{local_id}{ann["page_index"]}]'

            if ann["page_info"]:
                start_payload += f' {ann["page_info"]}\n'
            # elif ann["reference"]:
            #     start_payload += f' {ann["reference"]}\n'
            else:
                start_payload += "\n"
            only_start_ann = True
        elif ann["type"] == AnnType.topic:
            start_payload = f"{{{ann['work_id']}}}"
            only_start_ann = True
        elif ann["type"] == AnnType.sub_topic:
            start_payload = f"{{{ann['work_id']}}}"
            only_start_ann = True
        elif ann["type"] == AnnType.correction:
            start_payload = f"<{local_id}"
            end_payload = f',{ann["correction"]}>'
        elif ann["type"] == AnnType.archaic:
            start_payload = f"{{{local_id}"
            end_payload = f',{ann["modern"]}}}'
        elif ann["type"] == AnnType.peydurma:
            start_payload = f"#{local_id}"
            only_start_ann = True
        elif ann["type"] == AnnType.error_candidate:
            start_payload = f"[{local_id}"
            end_payload = "]"
        elif ann["type"] == AnnType.book_title:
            start_payload = f"<{local_id}k1"
            end_payload = ">"
        elif ann["type"] == AnnType.book_number:
            start_payload = f"<{local_id}k4"
            end_payload = ">"
        elif ann["type"] == AnnType.poti_title:
            start_payload = f"<{local_id}k2"
            end_payload = ">"
        elif ann["type"] == AnnType.author:
            start_payload = f"<{local_id}au"
            end_payload = ">"
        elif ann["type"] == AnnType.chapter:
            start_payload = f"<{local_id}k3"
            end_payload = ">"
        elif ann["type"] == AnnType.tsawa:
            start_payload = f"<{local_id}m"
            end_payload = "m>"
        elif ann["type"] == AnnType.citation:
            start_payload = f"<{local_id}g"
            end_payload = "g>"
        elif ann["type"] == AnnType.sabche:
            start_payload = f"<{local_id}q"
            end_payload = "q>"
        elif ann["type"] == AnnType.yigchung:
            start_payload = f"<{local_id}y"
            end_payload = "y>"

        start_cc, end_cc = self._get_adapted_span(ann["span"], vol_id)
        # start_cc -= 4
        self.add_chars(vol_id, start_cc, True, start_payload)
        if not only_start_ann:
            self.add_chars(vol_id, end_cc, False, end_payload)

    def serialize(self, output_path="./output/publication"):
        pecha_id = self.opf_path.stem
        self.apply_layers()
        results = self.get_result()
        vol2fn_manager = Vol2FnManager(self.get_meta_data())
        output_path = Path(output_path) / pecha_id
        output_path.mkdir(exist_ok=True, parents=True)
        print("[INFO] Creating HFML view")
        for vol_id, hfml_text in results.items():
            fn = vol2fn_manager.get_fn(vol_id)
            try:
                vol_hfml_fn = output_path / fn
            except Exception:
                vol_hfml_fn = output_path / vol_id
            print(f"[INFO]\t- saving {fn}")
            vol_hfml_fn.write_text(hfml_text)



import base64
import codecs
import csv
import glob
import json
import logging
import os
import re
import traceback
from contextlib import closing
from pathlib import Path

import requests
from tqdm import tqdm

from openpecha.buda.errors import Error
from openpecha.buda.openpecha_git import OpenpechaGit
from openpecha.serializers.rdf import Rdf


class OpenpechaManager:
    def __init__(self, cache_dir=str(Path.home() / ".cache" / "openpecha")):
        self.openpecha_api = "https://api.github.com/repos/OpenPecha"
        self.openpecha_git = "https://github.com/OpenPecha"
        self.cache_dir = cache_dir
        self.commits = None
        if not Path(cache_dir).is_dir():
            Path(cache_dir).mkdir(parents=True)

    def get_collection_blob_url(self):
        """
        Get the url for the README blob for the work collection.

        TODO: maybe a csv file should be fetched instead?

        TODO: we should use a local repo that we pull instead of the github api directly (less urgent)
        """
        catalog_url = f"{self.openpecha_api}/openpecha-catalog/git/trees/master"
        data = requests.get(url=catalog_url).json()  # getting the full repo
        try:
            tree_entry = [t for t in data["tree"] if t["path"] == "README.md"][
                0
            ]  # find the README to get its URL
            return tree_entry["url"]
        except IndexError:
            Error(
                IndexError,
                f"The README.md file not found at the root of the openpecha git\n```{traceback.format_exc()}```",
            )

    @staticmethod
    def get_blob_content(blob_url):
        """
        Get the blob content from the url provided in parameter
        """
        base64_content = requests.get(blob_url).json()["content"]

        return base64_content

    @staticmethod
    def decode_bas64_blob(base64_blob):
        """
         Decode the base64 encoded blob into UTF-8
        """
        decoded_list_data = base64.b64decode(base64_blob).decode("utf-8")

        return decoded_list_data

    @staticmethod
    def get_op_lname_column(decoded_blob):
        """
        Get the op_lname from the decoded blob and return a list of op_lname
        """
        op_lname = re.compile(r"\[P[0-9]{6}\]").findall(decoded_blob)

        return op_lname

    @staticmethod
    def cleanup_op_lname_list(collection):
        """
        Remove the brackets from [PXXXXXX] for every op_lname
        """
        op_lnames = [s.replace("[", "").replace("]", "") for s in collection]

        return op_lnames

    def get_list_of_pecha(self):
        """
        Getting the full list of poti based on the openpecha-catalog
        The catalog is too big to get the whole content through normal github API.
        It's necessary to get the blob through the Git Data API, for that we need the files
        hash or URL
        """
        collection_blob_url = self.get_collection_blob_url()
        if collection_blob_url:
            blob_content = self.get_blob_content(collection_blob_url)
            decoded_blob = self.decode_bas64_blob(blob_content)
            op_lname_column = self.get_op_lname_column(decoded_blob)
            cleaned_list = self.cleanup_op_lname_list(op_lname_column)

            return cleaned_list

    def fetch_all_pecha(self):
        """
        Getting all the poti from the list and either cloning them or pulling to the latest commit.
        All the repo are going to be stored in a local directory set by self.cache_dir
        """
        for lname in tqdm(
            self.get_list_of_pecha(), ascii=True, desc="Cloning or pulling the pecha"
        ):
            self.fetch_pecha(lname)

    def fetch_pecha(self, lname):
        """
        Getting one pecha from the list and either cloning them or pulling to the latest commit.
        """
        opg = OpenpechaGit(lname, cache_dir=self.cache_dir)
        try:
            opg.get_repo(dst_sync=True)
        except Exception:
            logging.error("error fetching %s", lname)

    def get_local_poti_info(self, get_commit=True):
        """
        Getting all the poti from the list and either cloning them or pulling to the latest commit.
        All the repo are going to be stored in a local directory set by self.cache_dir
        """
        res = {}
        for d in glob.glob(self.cache_dir + "/*"):
            lname = os.path.basename(d)
            if not lname.startswith("P"):
                continue
            res[lname] = {}
            if get_commit:
                opg = OpenpechaGit(lname, cache_dir=self.cache_dir)
                rev = opg.get_local_latest_commit()
                res[lname]["rev"] = rev
        return res

    @staticmethod
    def fetch_op_commits(ldspdibaseurl="http://ldspdi.bdrc.io/"):
        """
        Fetches the list of all openpecha commits on BUDA
        """
        res = {}
        headers = {"Accept": "text/csv"}
        params = {"format": "csv"}
        with closing(
            requests.get(
                ldspdibaseurl + "/query/table/OP_allCommits",
                stream=True,
                headers=headers,
                params=params,
            )
        ) as r:
            reader = csv.reader(codecs.iterdecode(r.iter_lines(), "utf-8"))
            for row in reader:
                if not row[0].startswith("http://purl.bdrc.io/resource/IE0OP"):
                    Error("store", "cannot interpret csv line starting with " + row[0])
                    continue
                res[row[0][34:]] = row[1]
        return res

    def get_buda_op_commits(self, ldspdibaseurl, force=False):
        if self.commits is not None and not force:
            return self.commits
        path = Path(self.cache_dir, "buda-commits.json")
        if path.is_file() and not force:
            with open(path) as json_file:
                self.commits = json.load(json_file)
                return self.commits
        commits = self.fetch_op_commits(ldspdibaseurl)
        with open(path, "w") as outfile:
            json.dump(commits, outfile)
        self.commits = commits
        return commits

    def set_commit(self, oplname, rev):
        if self.commits is None:
            logging.warn("updating op_commits without fetching commits first")
            return
        self.commits[oplname] = rev

    def write_commits(self):
        if self.commits is None:
            logging.warn("not writing None commits")
            return
        path = Path(self.cache_dir, "buda-commits.json")
        with open(path, "w") as outfile:
            json.dump(self.commits, outfile)

    @staticmethod
    def send_model_to_store(model, graphuri, storeurl):
        logging.info("sending %s to store", graphuri)
        ttlstr = model.serialize(format="turtle")
        headers = {"Content-Type": "text/turtle"}
        params = {"graph": graphuri}
        r = requests.put(storeurl, data=ttlstr, headers=headers, params=params)
        sc = r.status_code
        if (
            sc != requests.codes.ok
            and sc != requests.codes.created
            and sc != requests.codes.accepted
        ):
            Error("store", "The request to Fuseki returned code " + str(r.status_code))

    @staticmethod
    def write_model_debug(model, graphuri):
        lastslidx = graphuri.rfind("/")
        graphlname = graphuri[lastslidx + 1 :]
        fname = "/tmp/" + graphlname + ".ttl"
        model.serialize(destination=fname, format="turtle")

    def sync_cache_to_store(self, storeurl, ldspdibaseurl, force=False, idlist=None):
        buda_commits = {}
        if not force:
            buda_commits = self.get_buda_op_commits(ldspdibaseurl, force)
        for oplname, info in tqdm(
            self.get_local_poti_info(get_commit=(not force)).items(),
            ascii=True,
            desc="Converting into rdf",
        ):
            if idlist is not None and oplname not in idlist:
                continue
            if (
                force
                or (oplname not in buda_commits)
                or buda_commits[oplname] != info["rev"]
            ):
                # we need to sync this repo
                opgit = OpenpechaGit(oplname, cache_dir=self.cache_dir)
                op = opgit.get_openpecha()
                if op is None:
                    logging.error(
                        "skipping %s, can't get the openpecha object", oplname
                    )
                    continue
                if not op.is_ocr():
                    logging.info("skipping %s, not ocr", oplname)
                    continue
                rdf = Rdf(oplname, op)
                rdfgraph = rdf.get_graph()
                self.send_model_to_store(rdfgraph, str(rdf.graph_r), storeurl)
                # self.write_model_debug(rdfgraph, str(rdf.graph_r))
                self.set_commit(oplname, opgit.openpecharev)
            else:
                logging.info("skipping %s, already synced", oplname)
        self.write_commits()



import copy
import math

import diff_match_patch as dmp_module
import yaml


class Blupdate:
    """
    This class represents an update in the base layer. It is used to recompute the existing layers into the new base layer.

    When you want to update the base layer to a new text, initialize the class with the content of the old base layer and the new one.
    Then for each annotation in the layers, you can call get_updated_coord() with the coordinate of the annotation, it will return the
    updated coordinate that you can use to update the annotation.
    """

    def __init__(self, srcbl, dstbl, context_len=16):
        self.srcbl = srcbl
        self.dstbl = dstbl
        self.dmp = dmp_module.diff_match_patch()
        self.cctv = self.compute_cctv()
        self.context_len = context_len

    def compute_cctv(self):
        """
        Computes a cctv from self.srcbl to self.dstbl. This will take some effort but should be reasonable with DMP.

        The cctv should the a list of triples, each triple being:
          - char start in srcbl (int)
          - char end in srcbl (int)
          - char coord diff to the corresponding range in dstbl (int)


        It indicates the areas where no change has occured between srcbl and dstbl and where a simple character
        coordinate translation is possible without looking at the context.

        Here's an example:

        srcbl = abefghijkl
        dstbl = abcdefgkl

        the result will be:
        [
            (0,2,0),
            (2,5,2),
            (8,10,-1)
        ]

        meaning in plain English:
            - from character 0 to 2 in srcbl (meaning "ab"), no character coordinate translation is needed for the equivalent area in dstbl
            - from character 2 to 5 in srcbl (meaning "efg"), character coordinates in srcbl need to be added +2 in order to get their equivalent in dstbl
            - from character 8 to 10 in srcbl (meaning "kl"), character coordinates in srcbl need to be added -1 in order to get their equivalent in dstbl

        """
        diffs = self.dmp.diff_main(self.srcbl, self.dstbl)

        cctv = []

        # src_chunk is the chunk from source text, represented by mode 0 and -1.
        src_chunk_first_idx = 0
        src_chunk_last_idx = 0

        # dst_chunk is the chunk from destination text, represented by mode 0 and 1.
        dst_chunk_first_idx = 0
        dst_chunk_last_idx = 0

        for mode, chunk in diffs:
            if (
                mode == 0
            ):  # says the chunk is common, update the src_chunk and dst_chunk first and last indices.
                src_chunk_first_idx = src_chunk_last_idx
                dst_chunk_first_idx = dst_chunk_last_idx
                src_chunk_last_idx += len(chunk)
                dst_chunk_last_idx += len(chunk)
            elif (
                mode == -1
            ):  # says the chunk is from source text, update src_chunk first and last indices.
                src_chunk_first_idx = src_chunk_last_idx
                src_chunk_last_idx += len(chunk)
            else:  # mode == 1, says the chunks is from destination text, update dst_chunk first and last indices.
                dst_chunk_first_idx = dst_chunk_last_idx
                dst_chunk_last_idx += len(chunk)

            # get char coordinatte (cc) of the common chunk based on source and destination text.
            if mode == 0:
                src_cc = (src_chunk_first_idx, src_chunk_last_idx)
                dst_cc = (dst_chunk_first_idx, dst_chunk_last_idx)

                # contruct CCTV
                cctv.append((*src_cc, dst_cc[0] - src_cc[0]))

        return cctv

    def get_cctv_for_coord(self, srcblcoord):
        """
        Returns the character coordinate in dstbl for a character coordinate in srcbl according to the cctv. It returns two arguments:
          - the character coordinate translation, or an estimate if undefined
          - a boolean: true if the character coordinate is certain, false if it's an estimate

        In the relevant case, the estimate is computed by taking the average of the cctv value for the range before and for the range after.
        For example in the same example if we want the estimate for srcblcoord = 6, we see that for character coords [2-5] we have 2 and for [8-10] we have -1,
        the estimate will be 2+(-1) / 2 = 0.5, for which we take an int value: 1.

        So for a more complete example we would have:
          get_cctv_for_coord(3) == (2, true)
          get_cctv_for_coord(7) == (1, false)
          get_cctv_for_coord(9) == (-1, true)
        """
        prev_cct = 0
        result = None
        for cct in self.cctv:
            if srcblcoord > cct[0] and srcblcoord < cct[1]:  # at inner of the range
                result = (cct[2], True)
            elif (
                srcblcoord == cct[0] or srcblcoord == cct[1] - 1
            ):  # at side of the range
                result = (cct[2], False)
            elif srcblcoord < cct[0]:  # falls between ccts (two range)
                result = (math.ceil((prev_cct + cct[2]) / 2), False)

            if result:
                return result

            prev_cct = cct[2]

        if not result:
            return (-1, False)

    def get_context(self, srcblcoord):
        """
        This returns the left and right context of a character coordinate in srcbl, in the form of a tuple with two strings.
        The length of the context is set by self.context_len.

        format: 0a1b2e3f4g5h6i7

        For instance for contex_len = 4
           get_context(3) == ("abe", "fghi")
        """

        # check for left context size less than context_len
        if srcblcoord >= self.context_len:
            left_context = self.srcbl[srcblcoord - self.context_len : srcblcoord]
        else:
            left_context = self.srcbl[:srcblcoord]

        right_context = self.srcbl[srcblcoord : srcblcoord + self.context_len]
        return left_context, right_context

    def dmp_find(self, context, dstcoordestimate):
        """
        This function uses the dmp lib wizardry to get the a coordinate in dstbl that is:
          - around dstcoordestimate
          - approximately surrounded by context

        This is conceptually simple but might require the most time as it might require understanding the dmp code to
        hook into private functions.
        """
        match = self.dmp.match_main(self.dstbl, context[1], dstcoordestimate)
        return match

    def get_updated_with_dmp(self, srcblcoord, cct):
        """
        This returns the coordinate in dstbl corresponding to srcblcoord using the dmp methods, and an indication from the cctv.

        By convention, the function returns -1 when it is unable to compute the new coordinate.
        """
        if cct == -1:
            return cct
        context = self.get_context(srcblcoord)
        dstcoordestimate = srcblcoord + cct
        return self.dmp_find(context, dstcoordestimate)

    def get_updated_coord(self, srcblcoord):
        """
        This is the main function used to update annotations. Annotations have references to character coordinates in a specific base layer. This function
        allows them to get the corresponding coordinate in the new base layer.

        By convention, the return value -1 means that the function is unable to compute the new coordinate.
        """
        cctvforcoord = self.get_cctv_for_coord(srcblcoord)
        if cctvforcoord[1]:
            return srcblcoord + cctvforcoord[0]
        else:
            return self.get_updated_with_dmp(srcblcoord, cctvforcoord[0])

    def update_layer(self, layer_fn):
        """
        Update individual layer, in format {0: [0, 17], 1: [20, 40], ...}
        """
        anns = yaml.safe_load(layer_fn.open())
        for i, ann in anns.items():
            start_cc = ann[0]
            end_cc = ann[1]
            anns[i] = [self.get_updated_coord(start_cc), self.get_updated_coord(end_cc)]
        yaml.dump(anns, layer_fn.open("w"), default_flow_style=False)

    def update_annotations(self, opfpath):
        """
        Update all the layer annotations
        """
        for layer_fn in (opfpath / "layers").iterdir():
            self.update_layer(layer_fn)



import rdflib
from rdflib import BNode, Literal, URIRef
from rdflib.namespace import OWL, RDF, RDFS, SKOS, XSD, Namespace, NamespaceManager

rdf = RDF
rdfs = RDFS
bdr = Namespace("http://purl.bdrc.io/resource/")
bdo = Namespace("http://purl.bdrc.io/ontology/core/")
bdg = Namespace("http://purl.bdrc.io/graph/")
bda = Namespace("http://purl.bdrc.io/admindata/")
adm = Namespace("http://purl.bdrc.io/ontology/admin/")

nsm = NamespaceManager(rdflib.Graph())
nsm.bind("bdr", bdr)
nsm.bind("", bdo)
nsm.bind("bdg", bdg)
nsm.bind("bda", bda)
nsm.bind("adm", adm)
nsm.bind("rdf", rdf)
nsm.bind("rdfs", rdfs)



import gzip
import json
import math
import re
from pathlib import Path

import yaml

from ..utils import gzip_str
from .formatter import BaseFormatter
from .layers import *


class GoogleOCRFormatter(BaseFormatter):
    """
    OpenPecha Formatter for Google OCR JSON output of scanned pecha.
    """

    def __init__(self, output_path="./output", metadata=None):
        super().__init__(output_path, metadata)
        self.n_page_breaker_char = 3
        self.page_break = "\n" * self.n_page_breaker_char
        self.base_text = []

    def text_preprocess(self, text):

        return text

    def get_input(self, input_path):
        """
        load and return all jsons in the input_path.
        """
        for fn in sorted(list(input_path.iterdir())):
            if fn.name.split(".")[0] == "info":
                continue
            try:
                yield json.load(gzip.open(str(fn), "rb")), fn.name.split(".")[0]
            except GeneratorExit:
                return None, None
            except Exception:
                yield None, None

    def _get_page_index(self, n):
        page_sides = "ab"
        div = n / 2
        page_num = math.ceil(div)
        if div.is_integer():
            return f"{page_num}{page_sides[1]}"
        else:
            return f"{page_num}{page_sides[0]}"

    def format_layer(self, layers, base_id):
        # Format page annotation
        Pagination = Layer(self.get_unique_id(), "pagination")
        for pg, page_ref in zip(layers["pages"], layers["pages_ref"]):
            uuid = self.get_unique_id()
            span = Span(pg[0], pg[1])
            page = Page(span, page_index=self._get_page_index(pg[2]), page_ref=page_ref)
            Pagination["annotations"][uuid] = page

        result = {"pagination": Pagination}

        return result

    def _get_coord(self, vertices):
        coord = []
        for vertice in vertices:
            coord.append((vertice["x"], vertice["y"]))

        return coord

    def _get_page(self, response):
        try:
            page = response["textAnnotations"][0]
        except KeyError:
            return None, None

        text = page["description"]
        # vertices = page['boundingPoly']['vertices']  # get text box

        return text, None  # self._get_coord(vertices)

    def _get_lines(self, text, last_pg_end_idx, first_pg):
        lines = []
        line_breaks = [m.start() for m in re.finditer("\n", text)]

        start = last_pg_end_idx

        # increase the start idx with page_breaker_char for page greater than frist page.
        if not first_pg:
            start += self.n_page_breaker_char + 1
            line_breaks = list(map(lambda x: x + start, line_breaks))

        for line in line_breaks:
            lines.append((start, line - 1))  # skip new_line, which has 1 char length
            start += (line - start) + 1

        return lines, line

    def save_boundingPoly(self, response, path):
        def tlbr(vertices):
            return {"tl": vertices[0], "br": vertices[2]}

        boundingPolies = {}
        for ann in response["textAnnotations"]:
            boundingPolies[ann["description"]] = tlbr(ann["boundingPoly"]["vertices"])

        zipped_boundingPolies = gzip_str(json.dumps(boundingPolies))
        path.write_bytes(zipped_boundingPolies)

    def save_low_conf_char(self, response, path):
        char_idx = 0
        low_conf_chars = ""
        for page in response["fullTextAnnotation"]["pages"]:
            for block in page["blocks"]:
                for paragraph in block["paragraphs"]:
                    for word in paragraph["words"]:
                        for symbol in word["symbols"]:
                            if symbol.get("confidence", 1) < 0.9:
                                low_conf_chars += f'{symbol["text"]} {char_idx}\n'
                            char_idx += 1
        if low_conf_chars:
            path.write_bytes(gzip_str(low_conf_chars))

    def build_layers(self, responses, vol_name, base_id=None):
        if base_id:
            bounding_poly_vol_path = (
                self.dirs["release_path"] / "bounding_poly" / base_id
            )
            low_conf_char_vol_path = (
                self.dirs["release_path"] / "low_confidence_chars" / base_id
            )
            bounding_poly_vol_path.mkdir(parents=True, exist_ok=True)
            low_conf_char_vol_path.mkdir(parents=True, exist_ok=True)

        pages = []
        pages_ref = []
        last_pg_end_idx = 0
        for response, page_ref in responses:
            # extract page number, eg: I1PD901350083 -> 83
            n_pg_str = page_ref[len(vol_name) :]
            if "-" in page_ref:
                n_pg_str = page_ref.split("-")[-1]

            if n_pg_str and n_pg_str[-1].isalpha():
                n_pg_str = n_pg_str[:-1]

            try:
                n_pg = int(n_pg_str)
            except Exception:
                # TODO: fix later, collection all the cases as of now
                n_pg = 0  # dummy value

            # extract annotation
            if not response:
                print(f"[ERROR] Failed : {n_pg}")
                continue
            text, _ = self._get_page(response)

            # skip empty page (can be bad image)
            if not text:
                print(f"[ERROR] empty page {n_pg}")
                continue
            lines, last_pg_end_idx = self._get_lines(text, last_pg_end_idx, n_pg == 1)
            pages.append((lines[0][0], lines[-1][1], n_pg))
            pages_ref.append(response.get("image_link", page_ref))

            # create base_text
            self.base_text.append(text)

            if base_id:
                # save the boundingPoly to resources
                self.save_boundingPoly(
                    response, bounding_poly_vol_path / f"{n_pg:04}.json.gz"
                )
                # save low confident char and it's corresponding index
                self.save_low_conf_char(
                    response, low_conf_char_vol_path / f"{n_pg:04}.txt.gz"
                )

        result = {"pages": pages, "pages_ref": pages_ref}

        return result

    def get_base_text(self):
        base_text = f"{self.page_break}".join(self.base_text)
        self.base_text = []

        return base_text

    def get_metadata(self, work_id):
        import xml.etree.ElementTree as ET

        import requests
        from pyewts import pyewts

        converter = pyewts()
        query_url = "https://www.tbrc.org/xmldoc?rid={}"
        bdrc_metadata_url = query_url.format(work_id)
        r = requests.get(bdrc_metadata_url)

        try:
            root = ET.fromstring(r.content.decode("utf-8"))
        except Exception:
            metadata = {
                "id": f"opecha:{self.pecha_id}",
                "initial_creation_type": "ocr",
                "source_metadata": {
                    "id": f"{work_id}",
                    "title": "",
                    "volume": "",
                    "author": "",
                },
            }
            return metadata

        title_tag = root[0]
        author_tag = root.find("{http://www.tbrc.org/models/work#}creator")
        metadata = {
            "id": f"opecha:{self.pecha_id}",
            "initial_creation_type": "ocr",
            "source_metadata": {
                "id": f"bdr:{work_id}",
                "title": converter.toUnicode(title_tag.text),
                "volume": "",
                "author": converter.toUnicode(author_tag.text) if author_tag else "",
            },
        }

        return metadata

    def create_opf(self, input_path, pecha_id):
        input_path = Path(input_path)
        self._build_dirs(input_path, id_=pecha_id)

        self.dirs["release_path"] = self.dirs["opf_path"].parent / "releases"
        self.dirs["release_path"].mkdir(exist_ok=True, parents=True)

        for i, vol_path in enumerate(sorted(input_path.iterdir())):
            print(f"[INFO] Processing {input_path.name}-{vol_path.name} ...")
            base_id = f"v{i+1:03}"
            if (self.dirs["opf_path"] / "base" / f"{base_id}.txt").is_file():
                continue
            responses = self.get_input(vol_path)
            layers = self.build_layers(responses, vol_path.name, base_id)
            formatted_layers = self.format_layer(layers, base_id)
            base_text = self.get_base_text()

            # save base_text
            (self.dirs["opf_path"] / "base" / f"{base_id}.txt").write_text(base_text)

            # save layers
            vol_layer_path = self.dirs["layers_path"] / base_id
            vol_layer_path.mkdir(exist_ok=True)
            for layer, ann in formatted_layers.items():
                layer_fn = vol_layer_path / f"{layer}.yml"
                self.dump(ann, layer_fn)

        # create meta.yml
        meta_fn = self.dirs["opf_path"] / "meta.yml"
        self.dump(self.get_metadata(input_path.name), meta_fn)

        return self.dirs["opf_path"].parent


if __name__ == "__main__":
    formatter = GoogleOCRFormatter()
    formatter.create_opf("./tests/data/formatter/google_ocr/W3CN472", 300)



"""Utilities functions
"""
import gzip
import io
import shutil
from collections import defaultdict

from openpecha.github_utils import create_release


def gzip_str(string_):
    # taken from https://gist.github.com/Garrett-R/dc6f08fc1eab63f94d2cbb89cb61c33d
    out = io.BytesIO()

    with gzip.GzipFile(fileobj=out, mode="w") as fo:
        fo.write(string_.encode())

    bytes_obj = out.getvalue()
    return bytes_obj


def ocr_result_input(path):
    return path


def create_release_with_assets(path):
    asset_paths = []
    for asset_path in (path / "releases").iterdir():
        shutil.make_archive(asset_path, "zip", asset_path)
        asset_paths.append(f"{str(asset_path)}.zip")

    create_release(path.name, asset_paths=asset_paths)


class Vol2FnManager:
    def __init__(self, metadata):
        self.name = "vol2fn"
        self.vol_num = 0
        self.vol2fn = self._get_vol2fn(metadata)
        self.fn2vol = {fn: vol for vol, fn in self.vol2fn.items()}

    def _get_vol2fn(self, metadata):
        if self.name in metadata:
            return metadata[self.name]
        else:
            return defaultdict(dict)

    def get_fn(self, vol):
        return self.vol2fn.get(vol)

    def get_vol_id(self, fn):
        vol_id = self.fn2vol.get(fn)
        if vol_id:
            return vol_id
        else:
            self.vol_num += 1
            vol_id = f"v{self.vol_num:03}"
            self.vol2fn[vol_id] = fn
            return vol_id



"""catalog module contains all the functionalities necessary for managin
the catalog. Functonalities includes:
    - Creating opfs from input text
    - Assiging ID to the new opf
    - Updating the catalog with new opfs

"""

import os

import requests
import yaml

from openpecha.formatters import *
from openpecha.github_utils import create_file, create_readme, github_publish
from openpecha.utils import *

buildin_pipes = {
    "input": {"ocr_result_input": ocr_result_input},
    "release": {"create_release_with_assets": create_release_with_assets},
}


class CatalogManager:
    """Manages the catalog"""

    def __init__(
        self,
        pipes=None,
        formatter=None,
        layers=[],
        org="OpenPecha",
        token=os.environ.get("GITHUB_TOKEN"),
        not_include_files=["releases"],
        last_id_fn="last_id",
    ):
        self.org = org
        self.token = token
        self.repo_name = "catalog"
        self.batch_path = "data/batch.csv"
        self.last_id_path = f"data/{last_id_fn}"
        self.batch = []
        self.last_id = self._get_last_id()
        self.formatter = formatter
        self.layers = layers
        self.not_include_files = not_include_files
        self.pipes = pipes if pipes else buildin_pipes

    def _get_last_id(self):
        """returns the id assigin to last opf pecha"""
        last_id_url = f"https://raw.githubusercontent.com/{self.org}/{self.repo_name}/master/{self.last_id_path}"
        r = requests.get(last_id_url)
        return int(r.content.decode("utf-8").strip()[1:])

    def _add_id_url(self, row):
        id = row[0]
        row[0] = f"[{id}](https://github.com/{self.org}/{id})"
        return row

    def update(self):
        """Updates the catalog csv to have new opf-pechas metadata"""
        # update last_id
        content = self.batch[-1][0].strip()
        create_file(
            self.repo_name,
            self.last_id_path,
            content,
            "update last id of Pecha",
            update=True,
            org=self.org,
            token=self.token,
        )

        # update last_id
        self.last_id = int(content[1:])

        # create batch.csv
        content = (
            "\n".join([",".join(row) for row in map(self._add_id_url, self.batch)])
            + "\n"
        )
        create_file(
            self.repo_name,
            self.batch_path,
            content,
            "create new batch",
            org=self.org,
            token=self.token,
        )
        print("[INFO] Updated the catalog")

        # reset the batch
        self.batch = []

    def _get_catalog_metadata(self, meta_fn):
        metadata = yaml.safe_load(meta_fn.open())
        catalog_metadata = [
            metadata["id"].split(":")[-1],
            metadata["source_metadata"].get("title"),
            metadata["source_metadata"].get("subtitle"),
            metadata["source_metadata"].get("author"),
            metadata["source_metadata"].get("id"),
        ]
        self.batch.append(catalog_metadata)
        create_readme(metadata["source_metadata"], self.formatter.pecha_path)

    def format_and_publish(self, path):
        """Convert input pecha to opf-pecha with id assigined"""
        self.last_id += 1
        self.formatter.create_opf(path, self.last_id)
        self._get_catalog_metadata(self.formatter.meta_fn)
        github_publish(
            self.formatter.pecha_path,
            not_includes=self.not_include_files,
            layers=self.layers,
            org=self.org,
            token=self.token,
        )
        return self.formatter.pecha_path

    def add_ocr_item(self, path):
        self._process(path, "ocr_result_input", "create_release_with_assets")

    def add_hfml_item(self, path):
        self._process(path, "ocr_result_input")

    def _process(self, path, input_method, release_method=None):
        print("[INFO] Getting input")
        raw_pecha_path = self.pipes["input"][input_method](path)
        print("[INFO] Convert Pecha to OPF")
        opf_pecha_path = self.format_and_publish(raw_pecha_path)
        if release_method:
            print("[INFO] Release OPF pecha")
            self.pipes["release"][release_method](opf_pecha_path)



import logging
from pathlib import Path

import click
from bonltk.text_quality import non_words_ratio
from git import Repo
from tqdm import tqdm

from openpecha.buda.openpecha_manager import OpenpechaManager
from openpecha.catalog import config as catalog_config
from openpecha.catalog.filter import is_text_good_quality
from openpecha.catalog.storage import GithubBucket
from openpecha.formatters import *
from openpecha.serializers import *

OP_PATH = Path("./.openpecha")
config = {
    # Github
    "OP_CATALOG_URL": "https://raw.githubusercontent.com/OpenPoti/openpecha-catalog/master/data/catalog.csv",
    "OP_ORG": "https://github.com/OpenPecha",
    # Local
    "OP_DATA_PATH": (OP_PATH / "data").resolve(),
    "OP_CATALOG_PATH": (OP_PATH / "data" / "catalog.csv").resolve(),
    "CONFIG_PATH": (OP_PATH / "config").resolve(),
    "DATA_CONFIG_PATH": (OP_PATH / "config" / "data_config").resolve(),
}

ERROR = "[ERROR] {}"
INFO = "[INFO] {}"


def get_pecha_id(n):
    return f"P{int(n):06}"


@click.group()
def cli():
    pass


def create_config_dirs():
    config["OP_DATA_PATH"].mkdir(parents=True, exist_ok=True)
    config["CONFIG_PATH"].mkdir(parents=True, exist_ok=True)


def get_pecha(id, batch_path, layers):
    def _check_pecha(id=None, pechas=None, layer=None, pecha_list=None):
        if id not in pecha_list:
            if id in pechas:
                if layer:
                    if layer in pechas[id][-1].split("_"):
                        pecha_list.append(id)
                    else:
                        msg = f"{layer} layer is not found in {id}"
                        click.echo(ERROR.format(msg))
                else:
                    pecha_list.append(id)
            else:
                msg = f"{id} not found in OpenPecha catalog"
                click.echo(ERROR.format(msg))

    def _get_batch(batch_path):
        with Path(batch_path).open() as f:
            batch_ids = [line.strip() for line in f.readlines()]
        return batch_ids

    pecha_list = []

    # If filter by layers
    if layers:
        layers_name = [layer.strip() for layer in layers.split(",")]
        for layer in layers_name:
            batch_ids = None
            if id:
                _check_pecha(id=id, pechas=pechas, layer=layer, pecha_list=pecha_list)
            elif batch_path:
                if not batch_ids:
                    batch_ids = _get_batch(batch_path)
                for b_id in batch_ids:
                    _check_pecha(
                        id=b_id, pechas=pechas, layer=layer, pecha_list=pecha_list
                    )
            else:
                for p_id in pechas:
                    _check_pecha(
                        id=p_id, pechas=pechas, layer=layer, pecha_list=pecha_list
                    )
    else:
        if id:
            _check_pecha(id=id, pechas=pechas, pecha_list=pecha_list)
        elif batch_path:
            batch_ids = _get_batch(batch_path)
            for b_id in batch_ids:
                _check_pecha(id=b_id, pechas=pechas, pecha_list=pecha_list)
        else:
            for p_id in pechas:
                _check_pecha(id=p_id, pechas=pechas, pecha_list=pecha_list)

    return pecha_list


def download_pecha(pecha, out):
    # clone the repo
    pecha_url = f"{config['OP_ORG']}/{pecha}.git"
    pecha_path = config["OP_DATA_PATH"] / pecha
    if pecha_path.is_dir():  # if repo is already exits at local then try to pull
        repo = Repo(str(pecha_path))
        repo.heads["master"].checkout()
        repo.remotes.origin.pull()
    else:
        Repo.clone_from(pecha_url, str(pecha_path))


# Poti Download command
@cli.command()
@click.option("--number", "-n", help="Pecha number of pecha, for single pecha download")
@click.option(
    "--batch",
    "-b",
    help="path to text file containg list of names of \
                                     pecha in separate line. Poti batch download",
)
@click.option(
    "--filter",
    "-f",
    help="filter pecha by layer availability, specify \
                                     layer names in comma separated, eg: title,yigchung,..",
)
@click.option("--out", "-o", default="./pecha", help="directory to store all the pecha")
def download(**kwargs):
    """
    Command to download pecha.
    If id and batch options are not provided then it will download all the pecha.
    """
    pecha_id = get_pecha_id(kwargs["number"])

    # create config dirs
    create_config_dirs()

    # configure the data_path
    config["data"] = Path(kwargs["out"]).resolve()

    # get pecha
    # pechas = get_pecha(work_id, kwargs['batch'], kwargs['filter'])
    pechas = [pecha_id]

    # download the repo
    for pecha in tqdm(pechas):
        download_pecha(pecha, kwargs["out"])

    # save data_path in data_config
    config_path = config["DATA_CONFIG_PATH"]
    if not config_path.is_file():
        config_path.write_text(str(config["data"].resolve()))

    # print location of data
    msg = f"Downloaded {pecha_id} ... ok"
    click.echo(INFO.format(msg))


# OpenPecha Formatter
formatter_types = ["ocr", "hfml(default)", "tsadra"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(formatter_types), help="Type of formatter"
)
@click.option("--id", "-i", type=int, help="Id of the pecha")
@click.option("--output_path", "-o", help="output path to store opf pechas")
@click.argument("input_path")
def format(**kwargs):
    """
    Command to format pecha into opf
    """
    if kwargs["name"] == "ocr":
        formatter = GoogleOCRFormatter(kwargs["output_path"])
    elif kwargs["name"] == "tsadra":
        formatter = HFMLFormatter(kwargs["output_path"])
    else:
        formatter = HFMLFormatter(kwargs["output_path"])

    formatter.create_opf(kwargs["input_path"], kwargs["id"])


@cli.command()
@click.option("--cache-folder", "-c", help="path to the folder of the local cache")
@click.option("--idlist", "-l", help="comma-separated list of Openpecha IDs")
def pull_pechas(cache_folder, idlist):
    """
    Command to pull all the pechas in local cache
    """
    opm = None
    if cache_folder is not None:
        opm = OpenpechaManager(local_dir=cache_folder)
    else:
        opm = OpenpechaManager()
    if idlist is None:
        opm.fetch_all_pecha()
    else:
        opids = idlist.split(",")
        for opid in tqdm(opids, ascii=True, desc="Cloning or pulling the pecha"):
            opm.fetch_pecha(opid)


@cli.command()
@click.option("--cache-folder", "-c", help="path to the folder of the local cache")
@click.option("--store-uri", "-s", help="Fuseki URI", required=True)
@click.option(
    "--force",
    "-f",
    help="force upload even when commit match with triple store",
    is_flag=True,
)
@click.option(
    "--ldspdi-uri", "-u", help="lds-pdi URI", default="https://ldspdi.bdrc.io/"
)
@click.option("--idlist", "-l", help="comma-separated list of Openpecha IDs")
@click.option("--verbose", "-v", help="verbose", is_flag=True)
@click.option("--debug", "-d", help="debug", is_flag=True)
def cache_to_store(cache_folder, ldspdi_uri, store_uri, force, verbose, debug, idlist):
    """
    Update the cached version of synced commits
    """
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose:
        logging.basicConfig(level=logging.INFO)
    opm = None
    if cache_folder is not None:
        opm = OpenpechaManager(local_dir=cache_folder)
    else:
        opm = OpenpechaManager()
    opids = None
    if idlist is not None:
        opids = idlist.split(",")
    opm.sync_cache_to_store(store_uri, ldspdi_uri, force, opids)


export_types = ["hfml(default)", "epub"]


@cli.command()
@click.option(
    "--name", "-n", type=click.Choice(export_types), help="Type of export format"
)
@click.option("--opf_path", "-op")
@click.option("--output_path", "-o")
def export(**kwargs):
    """
    Command to export Pecha in epub
    """

    opf_path = kwargs["opf_path"]
    output_path = kwargs["output_path"]
    if not opf_path:
        opf_path = f'{config["OP_DATA_PATH"]}/{pecha_id}/{pecha_id}.opf'

    if kwargs["name"] == "epub":
        serializer = EpubSerializer(opf_path)
    else:
        serializer = HFMLSerializer(opf_path)
    serializer.serialize(kwargs["output_path"])


def _get_bucket(bucket_type, bucket_name, n):
    if bucket_type == "github":
        catalog_config.GITHUB_BUCKET_CONFIG["catalog"]["end"] = n
        return GithubBucket(bucket_name, config=catalog_config.GITHUB_BUCKET_CONFIG)


def _get_filter_strategy_caller(filter_strategy):
    if filter_strategy == "non_words_ratio":
        return non_words_ratio


def _save_text(text, output_path, parent_path, fn):
    pecha_path = Path(output_path) / parent_path
    pecha_path.mkdir(exist_ok=True)
    vol_path = pecha_path / fn
    vol_path.write_text(text)


@cli.command()
@click.option("--output_path", "-o", type=click.Path(exists=True), required=True)
@click.option("--bucket_type", "-bt", type=click.Choice(["github"]), default="github")
@click.option("--bucket_name", "-bn", type=str, default="OpenPecha")
@click.option(
    "--filter_strategy",
    "-fs",
    type=click.Choice(["non_words_ratio"]),
    default="non_words_ratio",
)
@click.option(
    "--threshold",
    "-th",
    type=float,
    default=0.8,
    help="Determines the quality of the text (1 being the highest and 0 being the lowest)",
)
@click.option("-n", type=int, default=1, help="number of pechas to download")
@click.option("--verbose", "-v", help="verbose", is_flag=True)
@click.option("--debug", "-d", help="debug", is_flag=True)
def corpus_download(
    output_path, bucket_type, bucket_name, filter_strategy, threshold, n, verbose, debug
):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose:
        logging.basicConfig(level=logging.INFO)

    bucket = _get_bucket(bucket_type, bucket_name, n)
    filter_strategy_caller = _get_filter_strategy_caller(filter_strategy)
    for pecha_id, base in bucket.get_all_pechas_base():
        for vol_base, vol_fn in base:
            if is_text_good_quality(
                vol_base, strategy=filter_strategy_caller, threshold=threshold
            ):
                _save_text(vol_base, output_path, pecha_id, vol_fn)



import re

from .serialize import Serialize


class SerializeFootNote(Serialize):
    """
    Serializes the foot-note for given text_id
    """

    def apply_annotation(self, vol_id, ann):
        payload = ""
        if ann["type"] == "pagination":
            payload = f'[{ann["page_index"]}] {ann["page_info"]}\n'
            start_cc, end_cc = self._get_adapted_span(ann["span"], vol_id)
            self.add_chars(vol_id, start_cc, True, payload)
        elif "peydurma" in ann["type"]:
            payload = f'<{ann["note"]}>'
            start_cc, end_cc = self._get_adapted_span(ann["span"], vol_id)
            self.add_chars(vol_id, end_cc, True, payload)
        elif ann["type"] == "correction":
            start_cc, end_cc = self._get_adapted_span(ann["span"], vol_id)
            error = self.base_layers[vol_id][start_cc : end_cc + 1]
            self.base_layers[vol_id] = (
                self.base_layers[vol_id][:start_cc]
                + ann["correction"]
                + self.base_layers[vol_id][end_cc + 1 :]
            )

            payload = f"<{error} སྡེ་དགེ།>"
            n_diff = len(ann["correction"]) - len(error)
            if not n_diff == 0:
                self.n_char_shifted.append((ann["span"]["start"], n_diff))
                start_cc, end_cc = self._get_adapted_span(ann["span"], vol_id)
            self.add_chars(vol_id, end_cc, False, payload)

    def generate_foot_notes(self, text):

        annotated_text = ""
        foot_notes = []
        notes = re.finditer(r"\<.*?\>", text)
        last_note_end_idx = 0
        for i, note in enumerate(notes):
            start, end = note.span(0)
            note_text = text[start:end]
            annotated_text += text[last_note_end_idx:start]
            foot_note_id = f"[^{i+1}K]"
            annotated_text += foot_note_id
            foot_notes.append(f"{foot_note_id}: {note_text[1:-1]}")
            last_note_end_idx = note.span(0)[1]
        else:
            annotated_text += text[last_note_end_idx:]

        return annotated_text + "\n\n" + "\n".join(foot_notes)

    def get_result(self):
        self.layers.pop()
        annotated_text, text_id = Serialize.get_result(self)
        return self.generate_foot_notes(annotated_text), text_id


if __name__ == "__main__":
    OPF_PECHA_PATH = "../openpecha-user/.openpecha/data/P000002/P000002.opf"

    serializer = SerializeFootNote(
        OPF_PECHA_PATH,
        text_id="D1794",
        layers=["correction", "peydurma-note", "pagination"],
    )
    serializer.apply_layers()
    annotated_text, text_id = serializer.get_result()
    print(annotated_text)



import re
from pathlib import Path

from setuptools import find_packages, setup


def read(fname):
    p = Path(__file__).parent / fname
    with p.open(encoding="utf-8") as f:
        return f.read()


def get_version(prop, project):
    project = Path(__file__).parent / project / "__init__.py"
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), project.read_text()
    )
    return result.group(1)


setup(
    name="openpecha",
    version=get_version("__version__", "openpecha"),
    author="Esukhia developers",
    author_email="esukhiadev@gmail.com",
    description="OpenPecha Toolkit allows state of the art for distributed standoff annotations on moving texts",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="Apache2",
    url="https://github.com/OpenPoti/openpecha-toolkit",
    packages=find_packages(),
    install_requires=[
        "Click==7.0",
        "diff-match-patch==20181111",
        "PyYAML>=5.0.0",
        "requests==2.22.0",
        "tqdm==4.35.0",
        "PyGithub==1.43.8",
        "GitPython==3.1.0",
        "bs4",
        "pyewts",
        "rdflib==5.0.0",
    ],
    python_requires=">=3.6",
    tests_require=["pytest"],
    entry_points={
        "console_scripts": [
            "openpecha=openpecha.cli:cli"
        ]  # command=package.module:function
    },
)



# TODO: use inheritance and abstract classes properly, see
# https://stackoverflow.com/questions/372042/difference-between-abstract-class-and-interface-in-python

from .serialize import Serialize


class SerializeMd(Serialize):
    """
    This class is serializes an opf file into MarkDown.

    Markdown markup for layers:
        title    --> # (Heading)
        yigchung --> *(italic)*
        tsawa    --> **(blod)**
        quotes   --> `(span)`
        sapche   --> [<sapche>](#) link
    """

    def apply_annotation(self, ann):
        """
        This applies an annotation.

        This is currently a dummy example.
        """
        if ann.type == "title":
            self.add_chars(ann.start_cc, True, "# ")
        elif ann.type == "yigchung":
            self.add_chars(ann.start_cc, True, "*")
            self.add_chars(ann.end_cc, False, "*")
        elif ann.type == "tsawa":
            self.add_chars(
                ann.start_cc, True, "**"
            )  # TODO: remove new line after ** from parser
            self.add_chars(ann.end_cc, False, "**")
        elif ann.type == "quotes":
            self.add_chars(ann.start_cc, True, "`")
            self.add_chars(ann.end_cc, False, "`")
        elif ann.type == "sapche":
            self.add_chars(ann.start_cc, True, "[")
            self.add_chars(ann.end_cc, False, "](#)")
